-----------------------------------------------------------------------------
-- NFS Shift 2 car import
-- select the folder of the car you want to import
-- known bugs: tires missing some faces, also need UV mapping manually
-- hierarchy is NOT used, but objects are correcly placed
-- the rest of the objects are correctly uv mapped, and all textures automatically loaded
-- to keep the hi-res version of the car, select everything containing *LODB and *LODC and delete them
-- some cars have more bodywork kits, you can select for example *kit04 objects and keep them or remove them.
-- to get the car materials, use the material picker. 
-- you can load the car livery textures in the "paint" material
-- many thanks to chipicao for his great help in research!
-- made by vagos21, script available at XenTaX forums (vagos21@yahoo.com)
------------------------------------------------------------------------------

struct theme(name, size, offset)
struct elmts(firstAttr,numAttrs,numChildren,firstChild,nextSibling)
struct attrs(ID,type,val,num_values,nextAttr)
struct colls(ID,unknown1,unknown2,unknown3)
struct amatrix(position,orientation)

matrix =#() --matrices array
matsarr = #() --materials array
mm --multi material. contains all car materials as sub-materials

macroScript SHIFT2_CAR_IMPORT category:"myscripts"
(
	undo off
	
	--select car folder
	modelpath = getSavePath caption:"Select extracted Shift 2 car folder"
	--modelpath="J:\extracted\vehicles\pagani_c9"
	print ("Importing car: " + modelpath as string)

	
	for i=1 to matsarr.count do (deleteItem matsarr 1)
	for i=1 to matrix.count do (deleteItem matrix 1)
	--delete mm
	
	readMaterials modelpath --load all materials
	readCarXML modelpath --load car
	
) --main

fn readCarXML modelpath =
(
	
	print "Importing meshes..."
	
	words = filterString modelpath "\\"
	lastname=words[words.count]
	filename=modelpath+"\\"+lastname+".vhf"

	-----------------------read matrices---------------------
	f = openFile filename
	skiptonextline f
	skiptonextline f
	words=filterstring (readline f) "\""
	num_matrices=words[8] as integer
	for i=1 to num_matrices do
	(
		skipToString f "MATRIX id="
		matrixline = readline f
		words = filterstring matrixline "\""
		l1 = filterstring words[3] " "
		pos = [ l1[1] as float ,l1[2] as float ,l1[3] as float ]
		l2 = filterstring words[5] " "
		orient = [ l2[1] as float ,l2[2] as float ,l2[3] as float ,l2[4] as float]
		matr = amatrix pos orient
		append matrix matr
	)
	close f
	
	------------------read  OBJECTS---------------------
	f = openFile filename
	while not (eof f) do
	(
		if (skipToString f "OBJECT")==undefined then exit
		
		words =filterstring (readline f) "\""
		name=words[2]
		matrixIdx = words[4] as integer + 1
		words =filterstring (readline f) "\""
		filepath = words[2]
		
		
		words = filterString modelpath "\\"
		texpath=""
		for i=1 to words.count-2 do
		texpath+= (words[i]+"\\")
		texpath+=filepath
		
		importPart texpath matrix[matrixIdx].position matrix[matrixIdx].orientation
	)
	
	close f
	
	print "Done!"

	
)

fn importPart filename position orientation=
(
	farr = #() --faces array
	varr = #() --vertex positions array
	UVWarr = #() --vertex UV coords array
	faceMatID = #() --material ID of face
	normarr = #() --vertex normals array
	hasnormals = false
	
	--open file
	f = fopen filename "rb"
	
	--jump first 8 bytes
	b1=readbyte f #unsigned
	b2=readbyte f #unsigned
	b3=readbyte f #unsigned
	b4=readbyte f #unsigned
	b5=readbyte f #unsigned
	b6=readbyte f #unsigned
	b7=readbyte f #unsigned
	b8=readbyte f #unsigned
	
	--read part name
	partName = (readString f)
--	print ("Loading part: " + partName)
	
	--jump remaining alignment zero bytes
	multiplier = (partName.count+1)/4
	if ( (mod (partName.count+1) 4) !=0 ) then
	(
		bytesToJump= ((multiplier+1)*4) - (partName.count+1)
		fseek f bytesToJump #seek_cur
	)
	
	--read number of vertices
	numVerts = readlong f #unsigned
	
	--read number of vertex properties
	numVertProps = readlong f #unsigned

	--read number of primitives
	numPrims = readlong f #unsigned
	
	--jump 10 unknown floats
	fseek f 40 #seek_cur
	
	
	if b5==1 then --skip bones info
	(
		numbones = readlong f #unsigned
		numchars = readlong f #unsigned
		fseek f numchars #seek_cur
		fseek f (numbones*48) #seek_cur
	)

	------------------------------read vertex properties---------------------------
	for prop=1 to numVertProps do
	(
		rec1 = readlong f #unsigned
		rec2 = readlong f #unsigned
		rec3 = readlong f #unsigned
--print (rec1 as string + " " + 	rec2 as string + " " + rec3 as string)	

		if (rec1==2 and rec2==0 and rec3==0) then --read positions
		(
			for v = 1 to numverts do
			(
				x=readfloat f
				y=readfloat f
				z=readfloat f
				avert=[x,z,y]
				append varr avert
			)
		)
		
		if (rec1==4 and rec2==6 and rec3==0) then --jump colors
		( fseek f (numverts*4) #seek_cur )
		
		if (rec1==2 and rec2==2 and rec3==0) then --read normals
		(
			hasnormals = true
			for n=1 to numVerts do
			(
				anorm = [readfloat f, readfloat f, readfloat f]
				append normarr anorm
			)
		)
		
		if (rec1==1 and rec2==3 and rec3==0) then --read UVs
		(
			for v = 1 to numverts do
			(
				aUVW = [readfloat f, -readfloat f, 0.0]
				append UVWarr aUVW
			)
		)
		
		if (rec1==0 and rec2==3 and rec3==3) then --jump zeroes..!?
		( fseek f (numverts*4) #seek_cur )
		
		if (rec1==2 and rec2==4 and rec3==0) then --read normals
		( fseek f (numverts*12) #seek_cur )
		
		if (rec1==2 and rec2==5 and rec3==0) then --jump ??
		( fseek f (numverts*12) #seek_cur )	
		
		if (rec1==1 and rec2==3 and rec3==2) then --jump UV set
		( fseek f (numverts*8) #seek_cur )
		
		if (rec1==1 and rec2==3 and rec3==1) then --jump UV set
		( fseek f (numverts*8) #seek_cur )
		
		if (rec1==2 and rec2==3 and rec3==0) then --jump ??
		( fseek f (numverts*12) #seek_cur )	
		
		if (rec1==2 and rec2==3 and rec3==1) then --jump ??
		( fseek f (numverts*12) #seek_cur )	
		
		if (rec1==1 and rec2==3 and rec3==4) then --jump UV set
		( fseek f (numverts*8) #seek_cur )
		
		if (rec1==1 and rec2==3 and rec3==3) then --jump UV set
		( fseek f (numverts*8) #seek_cur )
		
		if (rec1==4 and rec2==6 and rec3==1) then --jump colors
		( fseek f (numverts*4) #seek_cur )
		
		if (rec1==5 and rec2==8 and rec3==0) then --jump 4 bytes
		( fseek f (numverts*4) #seek_cur )
		
		if (rec1==3 and rec2==1 and rec3==0) then --jump ??
		( fseek f (numverts*16) #seek_cur )		
		
	) --end reading vertex properties
	
	-----------------------------read face properties-------------------------------
	for p=1 to numPrims do
	(
		--read materials file name
		matfilename = readstring f --material file name path
		matname = getFilenameFile matfilename --material name WITHOUT path

		--jump remaining alignment zero bytes
		multiplier = (matfilename.count+1)/4
		if ( (mod (matfilename.count+1) 4) !=0 ) then
		(
			bytesToJump= ((multiplier+1)*4) - (matfilename.count+1)
			fseek f bytesToJump #seek_cur
		)
		
		--jump a long of zeros
		fseek f 4 #seek_cur
		
		--read number of faces
		numFaces = readlong f
		
		--calculate face ID
		for i=1 to matsarr.count do
		(
			if (matchPattern matname pattern:matsarr[i].name) then
			(	
				for face=1 to numFaces do
				(
					append faceMatID i
				)
				exit
			)
		)
		
		--read faces
		for face=1 to numFaces do
		(
			ind1 = (readshort f #unsigned)+1
			ind2 = (readshort f #unsigned)+1
			ind3 = (readshort f #unsigned)+1
			aface = [ind1,ind3,ind2]
			append farr aface
		)
		
		--jump alignment zero bytes
		pos=ftell f
		multiplier = pos/4
		if ( (mod pos 4) !=0 ) then
		(
			bytesToJump= ((multiplier+1)*4) - pos
			fseek f bytesToJump #seek_cur
		)
		
		--read 2 shorts of some face index
		short1=readshort f #unsigned
		short2=readshort f #unsigned
		
		--jump 10 floats
		fseek f 40 #seek_cur
		
	) --read primitives
	
	--close file
	fclose f
	
	--build mesh	
	amesh = mesh vertices:varr faces:farr name:partname

	--set vertex normals
	--if hasnormals then
	--(
	--	for i = 1 to numVerts do
	--	(
	--		setNormal amesh i normarr[i]
	--	)
	--)
	
	--apply uvw tex coords to mesh
	applyuvw amesh UVWarr
	
	--apply multi-material to mesh
	amesh.material = mm
	
	--apply material IDs to mesh faces
	for i=1 to faceMatID.count do
	(
		setFaceMatID amesh i faceMatID[i]
	)
	
	--apply position and orientation
	amesh.pos = [position[1], position[3],position[2]]
	rotate amesh (quat -orientation[1] -orientation[3] -orientation[2] orientation[4])
	
) --importPart

fn readMaterials modelpath =
(
	print "Loading Materials and Textures..."
for filename in (getFilesRecursive modelpath "*.bmt") do
(
	--------------------------------READ MATERIAL FILE------------------------------
	f = fopen filename "rb"
	
	themesArr = #() --stuff at top
	elmt = #()
	attr = #()
	coll = #()
	numb = #()
	bool = #()
	strs = #()
	
	--read stuff at top of file
	for i=1 to 8 do
	(
		aName = ""
		for w=1 to 4 do (aName += (bit.intasChar(readByte f #unsigned)))
		aSize = readlong f #unsigned
		aOffset = readlong f #unsigned
		aZero = readlong f #unsigned
		atheme = theme aName aSize aOffset
		append themesArr atheme
	)
	
	--HEAD
	fseek f (themesArr[2].offset) #seek_set
	num_elts = readlong f #unsigned
	num_attr = readlong f #unsigned
	num_coll = readlong f #unsigned
	num_numb = readlong f #unsigned
	num_strs = readlong f #unsigned
	num_bool = readlong f #unsigned
	unknown1 = readlong f #unsigned
	
	--ELMT
	fseek f (themesArr[3].offset) #seek_set
	for i =1 to num_elts do
	(
		ID=readlong f #unsigned
		firstAttr=readlong f #unsigned
		numAttrs=readlong f #unsigned
		numChildren=readlong f #unsigned
		firstChild=readlong f #unsigned
		nextSibling=readlong f #unsigned
		numEltsSameType=readlong f #unsigned
		aelmt = elmts firstAttr numAttrs numChildren firstChild nextSibling
		append elmt aelmt
	)
	
	--ATTR
	fseek f (themesArr[4].offset) #seek_set
	for i =1 to  num_attr do
	(
		ID=readlong f #unsigned
		type=readlong f #unsigned
		val=readlong f #unsigned
		num_values=readlong f #unsigned
		nextAttr=readlong f #unsigned
		aattr = attrs ID type val num_values nextAttr
		append attr aattr
	)
	
	--COLL
	fseek f (themesArr[5].offset) #seek_set
	for i =1 to  num_coll do
	(
		ID=readlong f #unsigned
		unknown1=readlong f #unsigned
		unknown2=readlong f #unsigned
		unknown3=readlong f #unsigned
		acoll = colls ID unknown1 unknown2 unknown3
		append coll acoll
	)
	
	--NUMB
	fseek f (themesArr[6].offset) #seek_set
	for i =1 to  num_numb do
	(
		append numb (readfloat f)
	)
	
	--BOOL
	fseek f (themesArr[7].offset) #seek_set
	for i =1 to  4 do
	(
		append bool (readbyte f #unsigned)
	)
	
	--STRS
	fseek f (themesArr[8].offset) #seek_set
	for i =1 to  themesArr[8].size do
	(
		append strs (bit.intasChar(readByte f #unsigned))
	)
	
	--close file
	fclose f

	--print head 
	--print elmt 
	--print attr 
	--print coll 
	--print numb
	--print bool 
	--print strs
	
	words = filterString filename "\\"
	texpath=""
	for i=1 to words.count-3 do
		texpath+= (words[i]+"\\")
	
-----------------------------BUILD MATERIAL-----------------------------	
	amaterial = standard showInViewport:true Name: (getFilenameFile filename)
	previousattrib="" --keep last name for texture path'
	for i=1 to num_attr do
	(
		if attr[i].type==2 then --STRINGS
		(
			--take attribute name from characters array
			attribname=""
			startoffset=attr[i].val
			offadd=1
			do
			(
				achar =strs[startoffset+offadd]
				attribname += achar
				offadd += 1
			)while (achar != "")
			
			--take textures according to attrib name
			case attribname of
			(
				"diffuseTexture" :
				(
					adiffusemap = Bitmaptexture fileName: (texpath+previousattrib)
					amaterial.diffuseMap=(adiffusemap)
					aopacitymap = Bitmaptexture fileName: (texpath+previousattrib) monoOutput:1
					amaterial.opacityMap=(aopacitymap)
				)
				"specularTexture" :
				(
					aglossmap = Bitmaptexture fileName: (texpath+previousattrib)
					amaterial.specularLevelMap =(aglossmap)
				)
				"normalTexture" :
				(
					anormalmap = Bitmaptexture fileName: (texpath+previousattrib)
					amaterial.bumpMap=(anormalmap)
				)
			)
			
			if attr[i].val>0 then  previousattrib=attribname --keep last name for texture path'
			
		)
	)

	append matsarr amaterial
	
)
		
	mm = multimaterial numsubs:matsarr.count
	for i =1 to matsarr.count do
	(
		mm[i]=matsarr[i]
	)
	


	print "Done!"
) --readMaterials

fn readAlignedString f= --reads string aligned per 4 bytes
(
	--read string
	astring = (readString f)
	--jump remaining alignment bytes
	multiplier = (astring.count+1)/4
	if ( (mod (astring.count+1) 4) !=0 ) then
	(
		bytesToJump= ((multiplier+1)*4) - (astring.count+1)
		fseek f bytesToJump #seek_cur
	)
	return astring
) --readAlignedString

fn getFilesRecursive root pattern =
(
	dir_array = GetDirectories (root)
	for d in dir_array do
	join dir_array (GetDirectories (d+"*"))
	my_files = #()
	for f in dir_array do
	join my_files (getFiles (f + pattern))
	my_files 
) --getFilesRecursive

fn applyuvw anode auvarray	= --applies UVW to node
(
	if (auvarray.count > 0) then (
		anode.numtverts = anode.numverts
		buildtvfaces anode
		for vi = 1 to anode.numverts do (
			settvert anode vi auvarray[vi]
		)				
		allfaces = #()
		for fi = 1 to anode.numfaces do (
			crtface = getface anode fi
			append allfaces crtface
		)
		for fi = 1 to anode.numfaces do (
			settvface anode fi allfaces[fi]
		)
	)
)